import requests
import random
import geopandas as gpd
from shapely.geometry import shape
from .dataset import Dataset
from .utils import html_box

class Table(Dataset):
    """
    This is the main Table class.

    Parameters
    ----------
    id_hash: int
        An ID hash.
    attributes: dic
        A dictionary holding the attributes of a tabular dataset.
    server: str
        A string of the server URL.
    """
    def __init__(self, id_hash=None, attributes=None, server='https://api.resourcewatch.org'):
        super().__init__(id_hash=id_hash, attributes=attributes, server=server)

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        return f"Table {self.id} {self.attributes['name']})"

    def fetch_query(self, sql):
        """
        Forms a base query and returns data
        """
        table_name = self.attributes.get('tableName', 'data')
        sql = sql.replace('FROM data', f'FROM {table_name}')
        try:
            url = (f'{self.server}/v1/query/{self.id}?sql={sql}')
            r = requests.get(url)
            if r.status_code == 200:
                print(r.url)
                response_data = r.json().get('data')
                for d in response_data:
                    if d.get('the_geom', None):
                        d['geometry'] = shape(d['the_geom'])
                return response_data
            else:
                raise ValueError(f'Unable to get table {self.id} from {r.url}')
        except:
            raise ValueError(f'Unable to get table {self.id} from {r.url}')

    def head(self, n=5):
        """
        Returns a table as a GeoPandas GeoDataframe from a Vizzuality API using the query endpoint.

        Parameters
        ----------
        n: int
            Number of rows in the head.
        decode_geom: bool
            A flag to decode geometries into geom objects.
        """
        sql = f'SELECT * FROM data LIMIT {n}'
        response_data = self.fetch_query(sql=sql)
        try:
            gdf = gpd.GeoDataFrame(response_data)
            if 'geometry' in gdf:
                gdf = gdf.set_geometry('geometry')
            return gdf
        except:
            raise ValueError(f'Unable to get table {self.id}')

    def query(self, sql=None):
        """
        Return an SQL query as a valid dataframe object.

        Parameters
        ----------
        sql: str
            A valid SQL query e.g. 'SELECT * FROM data LIMIT 5'
        decode_geom: bool
            A flag to decode geometries into geom objects.
        """
        if not sql: sql = 'SELECT * FROM data LIMIT 5'
        if type(sql) != str:
            raise ValueError('SQL query should be passed as a string.')
        response_data = self.fetch_query(sql=sql)
        try:
            gdf = gpd.GeoDataFrame(response_data)
            if 'geometry' in gdf:
                gdf = gdf.set_geometry('geometry')
            return gdf
        except:
            raise ValueError(f'Unable to query table {self.id} with {sql}')


from .lmipy import Vocabulary, Metadata, Widget
from .layer import Layer
from .image import Image
from .imageCollection import ImageCollection
from .dataset import Dataset
from .geometry import Geometry
from .collection import Collection
from .table import Table
from pkg_resources import get_distribution

__version__ = get_distribution('LMIPy').version


import json
from shapely.geometry import mapping, shape, box

def html_box(item):
    """Returns an HTML block with template strings filled-in based on item attributes."""
    is_layer = str(type(item)) == "<class 'LMIPy.layer.Layer'>"
    is_dataset = str(type(item)) == "<class 'LMIPy.dataset.Dataset'>"
    is_table = str(type(item)) == "<class 'LMIPy.table.Table'>"
    is_widget = str(type(item)) == "<class 'LMIPy.lmipy.Widget'>"
    is_geometry = str(type(item)) == "<class 'LMIPy.geometry.Geometry'>"
    is_image = str(type(item)) == "<class 'LMIPy.image.Image'>"
    needs_widgets_and_co = server_uses_widgets(item.server)
    if needs_widgets_and_co:
        site_link = "<a href='https://resourcewatch.org/' target='_blank'>"
        site_logo = "<img class='itemThumbnail' src='https://resourcewatch.org/static/images/logo-embed.png'>"
    else:
        site_link = "<a href='https://skydipper.com/' target='_blank'>"
        site_logo = "<img class='itemThumbnail' src='https://skydipper.com/images/logo.png'>"
    if is_layer:
        kind_of_item = 'Layer'
        if needs_widgets_and_co:
            url_link = f'{item.server}/v1/layer/{item.id}?includes=vocabulary,metadata'
        else:
            url_link = f'{item.server}/v1/layer/{item.id}?includes=metadata'
    elif is_dataset:
        kind_of_item = 'Dataset'
        if needs_widgets_and_co:
            url_link = f'{item.server}/v1/dataset/{item.id}?includes=vocabulary,metadata,layer,widget'
        else:
            url_link = f'{item.server}/v1/dataset/{item.id}?includes=metadata,layer'
    elif is_table:
        kind_of_item = 'Table'
        if needs_widgets_and_co:
            url_link = f'{item.server}/v1/dataset/{item.id}?includes=vocabulary,metadata,layer,widget'
        else:
            url_link = f'{item.server}/v1/dataset/{item.id}?includes=metadata,layer'
    elif is_image:
        if item.type in ['Classified Image', 'Composite Image']:
            instrument = item.type
        else:
            instrument = item.instrument
        html_string = ("<div class='item_container' style='height: auto; overflow: hidden; border: 3px solid #2BA4A0;"
                    "border-radius: 2px; background: #fff; line-height: 1.21429em; padding: 10px;''>"
                    "<div class='item_left' style='width: 100px; height: 100px; float: left; padding-right:10px''>"
                    f"<a href='{item.thumb_url}' target='_blank'>"
                    f"<img class='itemThumbnail' src='{item.thumb_url}'>"
                    "</a></div><div class='item_right' style='float: none; width: auto; hidden;padding-left: 10px; overflow: hidden;''>"
                    f"<b>Image Source</b>: {instrument} </br>"
                    f"<b>Datetime</b>: {item.date_time} </br>"
                    f"<b>Cloud score </b>: {item.cloud_score} </br>"
                    " </div> </div>")
        return html_string
    elif is_widget:
        kind_of_item = 'Widget'
        url_link = f'{item.server}/v1/widget/{item.id}'
    elif is_geometry:
        kind_of_item = 'Geometry'
        url_link = f'{item.server}/v1/geostore/{item.id}'
        html_string = ("<div class='item_container' style='height: auto; overflow: hidden; border: 3px solid #2bA4A0;"
            "border-radius: 2px; background: #fff; line-height: 1.21429em; padding: 10px;''>"
            "<div class='item_left' style='width: 210px; float: left;''>"
            f"{site_link}"
            f"{site_logo}"
            "</a></div><div class='item_right' style='float: none; width: auto; hidden;padding-left: 10px; overflow: hidden;''>"
            f"<b>Geometry id</b>: <a href={url_link} target='_blank'>{item.id}</a></br>")
        for k,v in item.attributes.get('info').items():
            if v and k != 'simplifyThresh':
                html_string += f"<br><i>{k}: {v}</i>"
        html_string += (""
            " </div> </div>")
        return html_string
    else:
        kind_of_item = 'Unknown'
        url_link = None
    table_statement = f"Data source {item.attributes.get('provider')}"
    if item.attributes.get('connectorUrl') and item.attributes.get('provider') == "cartodb":
        table_statement = (f"Carto table: <a href={item.attributes.get('connectorUrl')}"
                           " target='_blank'>"
                           f"{item.attributes.get('tableName')}"
                           "</a>"
                          )
    if item.attributes.get('connectorUrl') and item.attributes.get('provider') == "csv":
        table_statement = (f"CSV Table: <a href={item.attributes.get('connectorUrl')}"
                           " target='_blank'>"
                           f"{item.attributes.get('tableName')}"
                           "</a>"
                          )
    if item.attributes.get('provider') == 'gee':
        table_statement = (f"GEE asset: <a href='https://code.earthengine.google.com/asset='"
                           f"{item.attributes.get('tableName')} target='_blank'>"
                           f"{item.attributes.get('tableName')}"
                           "</a>"
                          )
    html = ("<div class='item_container' style='height: auto; overflow: hidden; border: 3px solid #2BA4A0;"
            "border-radius: 2px; background: #fff; line-height: 1.21429em; padding: 10px;''>"
            "<div class='item_left' style='width: 210px; float: left;''>"
            f"{site_link}"
            f"{site_logo}"
            "</a></div><div class='item_right' style='float: none; width: auto; hidden;padding-left: 10px; overflow: hidden;''>"
            f"<a href={url_link} target='_blank'>"
            f"<b>{item.attributes.get('name')}</b>"
            "</a>"
            f"<br> {table_statement} | {kind_of_item} in {', '.join(item.attributes.get('application')).upper()}."
            f"<br>Last Modified: {item.attributes.get('updatedAt')}"
            f"<br><a href='{item.server}/v1/fields/{item.id}' target='_blank'>Fields</a>"
            f" Connector: {item.attributes.get('provider')}"
            f" | Published: {item.attributes.get('published')}"
            " </div> </div>")
    return html

def show_image_collection(item, i):
    html_string = ("<div class='item_container' style='height: auto; overflow: hidden; border: 3px solid #2BA4A0;"
                "border-radius: 2px; background: #fff; line-height: 1.21429em; padding: 10px;''>"
                "<div class='item_left' style='width: 100px; height: 100px; hidden;padding-left: 10px; float: left; padding-right:10px''>"
                f"<a href='{item.get('thumb_url')}' target='_blank'>"
                f"<img class='itemThumbnail' src='{item.get('thumb_url')}'>"
                "</a></div><div class='item_right' style='float: none; hidden;padding-left: 10px; width: auto; overflow: hidden;''>"
                f"<b>Image Source</b>: {item.get('instrument')} </br>"
                f"<b>Datetime</b>: {item.get('date_time')} </br>"
                f"<b>Cloud score </b>: {item.get('cloud_score')} </br>"
                " </div> </div>")
    return html_string

def show(item, i):
    """Returns an HTML block with template strings filled-in based on item attributes."""
    is_layer = item['type'] == 'Layer'
    is_dataset = item['type'] == 'Dataset'
    is_table = item['type'] == 'Table'
    is_widget = item['type'] == 'Widget'
    server = item['server']
    item_id = item['id']
    attributes = item['attributes']
    uses_widgets = server_uses_widgets(server)
    if is_layer:
        kind_of_item = 'Layer'
        if uses_widgets:
            url_link = f'{server}/v1/layer/{item_id}?includes=vocabulary,metadata'
        else:
            url_link = f'{server}/v1/layer/{item_id}?includes=metadata'
    elif is_dataset:
        kind_of_item = 'Dataset'
        if uses_widgets:
            url_link = f'{server}/v1/dataset/{item_id}?includes=vocabulary,metadata,layer,widget'
        else:
            url_link = f'{server}/v1/dataset/{item_id}?includes=metadata,layer'
    elif is_table:
        kind_of_item = 'Table'
        if uses_widgets:
            url_link = f'{server}/v1/dataset/{item_id}?includes=vocabulary,metadata,layer,widget'
        else:
            url_link = f'{server}/v1/dataset/{item_id}?includes=metadata,layer'
    elif is_widget:
        kind_of_item = 'Table'
        url_link = f'{server}/v1/widget/{item_id}'
    else:
        kind_of_item = 'Unknown'
        url_link = None
    table_statement = f"Data source {attributes.get('provider')}"
    if attributes.get('connectorUrl') and attributes.get('provider') == "cartodb":
        table_statement = (f"Carto table: <a href={attributes.get('connectorUrl')}"
                           " target='_blank'>"
                           f"{attributes.get('tableName')}"
                           "</a>"
                          )
    if attributes.get('connectorUrl') and attributes.get('provider') == "csv":
        table_statement = (f"CSV Table: <a href={attributes.get('connectorUrl')}"
                           " target='_blank'>"
                           f"{attributes.get('tableName')}"
                           "</a>"
                          )
    if attributes.get('provider') == 'gee':
        table_statement = (f"GEE asset: <a href='https://code.earthengine.google.com/asset='"
                           f"{attributes.get('tableName')} target='_blank'>"
                           f"{attributes.get('tableName')}"
                           "</a>"
                          )
    if uses_widgets:
        site_link = "<a href='https://resourcewatch.org/' target='_blank'>"
        site_logo = "<img class='itemThumbnail' src='https://resourcewatch.org/static/images/logo-embed.png'>"
    else:
        site_link = "<a href='https://skydipper.com/' target='_blank'>"
        site_logo = "<img class='itemThumbnail' src='https://skydipper.com/images/logo.png'>"
    html = ("<div class='item_container' style='height: auto; overflow: hidden; border: 3px solid #2BA4A0;"
            "border-radius: 2px; background: #fff; line-height: 1.21429em; padding: 10px;''>"
            "<div class='item_left' style='width: 210px;  float: left;''>"
            f"{site_link}"
            f"{site_logo}"
            "</a></div><div class='item_right' style='float: none; width: auto; hidden;padding-left: 10px; overflow: hidden;''>"
            f"<b>{i}. </b>"
            f"<a href={url_link} target='_blank'>"
            f"<b>{attributes.get('name')}</b>"
            "</a>"
            f"<br> {table_statement} | {kind_of_item} in {', '.join(attributes.get('application')).upper()}."
            f"<br>Last Modified: {attributes.get('updatedAt')}"
            f"<br><a href='{server}/v1/fields/{item_id}' target='_blank'>Fields</a>"
            f" | Connector: {attributes.get('provider')}"
            f" | Published: {attributes.get('published')}"
            " </div> </div>")
    return html

def create_class(item):
    from .dataset import Dataset
    from .table import Table
    from .layer import Layer
    from .lmipy import Widget
    from .image import Image
    if item['type'] == 'Table':
        return Table(id_hash = item.get('id'), server = item.get('server'))
    elif item['type'] == 'Dataset':
        return Dataset(id_hash = item.get('id'), server = item.get('server'))
    elif item['type'] == 'Layer':
        return Layer(id_hash = item.get('id'), server = item.get('server'))
    elif item['type'] == 'Widget':
        return Widget(id_hash = item.get('id'), attributes=item.get('attributes'), server = item.get('server'))
    elif item['type'] == 'Image':
        return Image(**item)

def flatten_list(nested_list):
    if len(nested_list) > 0:
        return [item for sublist in nested_list for item in sublist]
    else:
        return []

def get_geojson_string(geom):
    coords = geom.get('coordinates', None)
    if coords and not any(isinstance(i, list) for i in coords[0]):
        geom['coordinates'] = [coords]
    feat_col = {"type": "FeatureCollection", "features": [{"type": "Feature", "properties": {}, "geometry": geom}]}
    return json.dumps(feat_col)

def parse_filters(filter_objects):
    filter_whitelist = ['connectorType', 'provider', 'status', 'published', 'protected', 'geoInfo']
    return_string = ''
    error_string = ''
    if filter_objects:
        for k, v in filter_objects.items():
            if k in filter_whitelist:
                return_string += f'{k}={v}&'
            else:
                error_string += f' {k},'
        if error_string:
            print(f'Unable to filter by{error_string[:-1]}.')
        return return_string
    return ''

def sldDump(sldObj):
    """
    Creates valid SldStyle string from an object.
    """
    sld_type = sldObj.get('type', None)
    extended = str(sldObj.get('extended', 'false')).lower()
    items = sldObj.get('items', None)
    sld_attr = {'color': None, 'label': None, 'quantity': None, 'opacity': None}

    if sld_type not in ['linear', 'ramp', 'gradient', 'intervals', 'values']:
        print('Unable to create SldStyle. Type must be in "linear", "ramp", "gradient", "intervals", "values".')
        return None
    if items:
        sld_str = f'<RasterSymbolizer> <ColorMap type="{sld_type}" extended="{extended}"> '
        for item in items:
            sld_str += f'<ColorMapEntry '
            for k in sld_attr.keys():
                if item.get(k, None): sld_str += f'{k}="{item[k]}" '
            sld_str += '/> + '
    return sld_str + "</ColorMap> </RasterSymbolizer>"

def sldParse(sld_str):
    """
    Builds a dictionary from an SldStyle string.
    """
    sld_str = sld_str.replace("'", '"').replace('\"', '"')
    keys = ['color', 'label', 'quantity', 'opacity']
    items = [el.strip() for el in sld_str.split('ColorMapEntry') if '<RasterSymbolizer>' not in el]
    sld_items = []
    for i in items:
        tmp = {}
        for k in keys:
            v = find_between(i, f'{k}="', '"')
            if v: tmp[k] = v
        sld_items.append(tmp)
    return {
        'type': find_between(sld_str, 'type="', '"'),
        'extended': find_between(sld_str, 'extended="', '"'),
        'items': sld_items
    }

def find_between(s, first, last):
    try:
        start = s.index(first) + len(first)
        end = s.index(last, start)
        return s[start:end]
    except ValueError:
        return ""

def nested_set(dic, keys, value):
    for key in keys[:-1]:
        dic = dic.setdefault(key, {})
    dic[keys[-1]] = value

def server_uses_widgets(server):
    """
    Does the server currently set use Widget objects? Response gives True if it does, false if not.
    """
    uses_widgets = ['https://api.resourcewatch.org','https://staging-api.globalforestwatch.org']
    if any(server in s for s in uses_widgets):
        return True
    else:
        return False


def create_grid(feature, N=3, as_shp=False):
    """
    Breaks down a feature into an N x N grid and returns a new list of gridded features.
    :param feature:  geojson feature
    :param N: Grid dimension
    :param as_shp: Bool. If True returns a list of shapely objects
    :return: list of features
    """
    geom = feature.get('geometry', None)
    feature_shp = shape(geom)
    if not geom: return None

    # Get bounds of grid
    lon_end,lat_start,lon_start,lat_end = shape(geom).bounds

    num_cells = N * N
    lon_edge = (lon_end - lon_start) / (num_cells ** 0.5)
    lat_edge = (lat_end - lat_start) / (num_cells ** 0.5)

    # Generate grid over feature
    polys = []
    lon = lon_start
    for i in range(0, N):
        x1 = lon
        x2 = lon + lon_edge
        lon += lon_edge

        lat = lat_start
        for j in range(0, N):
            y1 = lat
            y2 = lat + lat_edge
            lat += lat_edge

            polygon = box(x1, y1, x2, y2)
            polys.append(polygon)

    # Intersects grid against feature
    intersected_feats = []
    for p in polys:
        is_intersect = p.intersects(feature_shp)
        if is_intersect:
            intersection = p.intersection(feature_shp)
            intersected_feats.append(intersection)

    if as_shp: return intersected_feats

    feats = []
    for f in intersected_feats:
        feat = mapping(f)
        feats.append({'type': 'Feature', 'properties': {}, 'geometry': json.dumps(feat)})

    return feats


import requests
import random
import os
import json
import datetime
from tqdm import tqdm
from .dataset import Dataset
from .table import Table
from .layer import Layer
from .utils import create_class, show, flatten_list, parse_filters, server_uses_widgets

class Collection:
    """
    Returns a list of objects from a server

    This function searches all avaiable layers or dataset entries within user specified limits and returns a list.
    of objects.

    Parameters
    ----------
    app: list
        A list of string IDs of applications to search, e.g. [‘gfw’, ‘rw’]
    limit: int
        Maximum number of items to return
    order: str
        Field to order items by, e.g. ’date’
    sort: str
        Rule to sort items by, either ascending (’asc’) or descending ('desc')
    search: str
        String to search records by, e.g. ’Forest loss’
    object_type: list
        A list of strings of object types to search, e.g. [‘dataset’, ‘layer’]
    filters: dict
        A dictionary of filter key, value pairs e.g. {'provider', 'gee'}
        Possible search keys: 'connectorType', 'provider', 'status', 'published', 'protected', 'geoInfo'.
    """
    def __init__(self, search='', app=['gfw','rw'], env='production', limit=1000, order='name', sort='desc',
                 object_type=['dataset', 'layer','table', 'widget'], server='https://api.resourcewatch.org',
                 filters=None, mapbox_token=None):
        self.search = [search.lower()] + search.lower().strip().split(' ')
        self.server = server
        self.app = ",".join(app)
        self.env = env
        self.limit = limit
        self.order = order
        self.sort = sort
        self.filters = filters
        self.mapbox_token = mapbox_token
        self.object_type = object_type
        self.collection = self.get_collection()
        self.iter_position = 0

    def _repr_html_(self):
        str_html = ""
        for n, c in enumerate(self.collection):
            str_html += show(c, n)
            if n < len(self.collection)-1:
                str_html += '<p></p>'
        return str_html

    def __repr__(self):
        rep_string = "["
        for n, c in enumerate(self.collection):
            rep_string += str(f"{n}. {c['type']} {c['id']} {c['attributes']['name']}")
            if n < len(self.collection)-1:
                rep_string += ',\n '
        rep_string += ']'
        return rep_string

    def __iter__(self):
        return self

    def __next__(self):
        if self.iter_position >= len(self.collection):
            self.iter_position = 0
            raise StopIteration
        else:
            self.iter_position += 1
            return self.collection[self.iter_position - 1]

    def __getitem__(self, key):
        items = self.collection[key]
        if type(items) == list:
            return [create_class(item) for item in items]
        else:
            return create_class(items)

    def __len__(self):
        return len(self.collection)

    def get_collection(self):
        """
        Getter for the a collection object. In this case dataset and layers
        are the objects in the API. I.e. tables are a dataset type.
        """
        datasets = self.get_entities()
        layers = []
        # for d in datasets:
        #     tmp_atts = d.get('attributes', None)
        #     layers = tmp_atts.get('layer', None)
        # layers = flatten_list(layers)
        layers = flatten_list([d.get('attributes').get('layer') for d in datasets])
        if server_uses_widgets(server=self.server):
            widgets = flatten_list([d.get('attributes').get('widget') for d in datasets])
        else:
            widgets = None
        response_list = []
        if 'layer' in self.object_type:
            _ = [response_list.append(l) for l in layers]
        if 'dataset' in self.object_type or 'table' in self.object_type:
            _ = [response_list.append(d) for d in datasets]
        if 'widget' in self.object_type:
            _ = [response_list.append(w) for w in widgets]
        filtered_list = self.filter_results(response_list)
        ordered_list = self.order_results(filtered_list)
        return ordered_list

    def get_entities(self):
        hash = random.getrandbits(16)
        filter_string = parse_filters(self.filters)
        if server_uses_widgets(server=self.server):
            url = (f'{self.server}/v1/dataset?app={self.app}&env={self.env}&{filter_string}'
                   f'includes=layer,vocabulary,metadata,widget&page[size]=1000&hash={hash}')
        else:
            url = (f'{self.server}/v1/dataset?app={self.app}&env={self.env}&{filter_string}'
                   f'includes=layer,metadata&page[size]=1000&hash={hash}')
        r = requests.get(url)
        response_list = r.json().get('data', None)
        if not response_list:
            raise ValueError('No items found')
        return response_list

    def filter_results(self, response_list):
        """Search by a list of strings to return a filtered list of Dataset or Layer objects"""
        filtered_response = []
        collection = []
        for item in response_list:
            return_layers = 'layer' in self.object_type
            return_datasets = 'dataset' in self.object_type
            return_tables = 'dataset' in self.object_type
            return_widgets = 'widget' in self.object_type
            #print(f"Filter: {type(item)}, {item}")
            if type(item) == dict:
                tmp_atts = item.get('attributes')
                name = tmp_atts.get('name', None)
                description = tmp_atts.get('description', None)
                slug = tmp_atts.get('slug', None)
            else:
                name = None
                description = None
                slug = None
            found = []
            if description:
                description = description.lower()
                found.append(any([s in description for s in self.search]))
            if name:
                name = name.lower()
                found.append(any([s in name for s in self.search]))
            if slug:
                slug = slug.lower().split('_')
                found.append(any([s in slug for s in self.search]))
            if any(found):
                if len(filtered_response) < self.limit:
                    filtered_response.append(item)
                if item.get('type') == 'dataset' and item.get('attributes').get('provider') in ['csv', 'json'] and return_tables:
                    collection.append({'type': 'Table','id': item.get('id'), 'attributes': item.get('attributes'), 'server': self.server})
                elif item.get('type') == 'dataset' and item.get('attributes').get('provider') != ['csv','json'] and return_datasets:
                    collection.append({'type': 'Dataset','id': item.get('id'), 'attributes': item.get('attributes'), 'server': self.server})
                if item.get('type') == 'layer' and return_layers:
                    collection.append({'type': 'Layer', 'id': item.get('id'), 'attributes': item.get('attributes'), 'server': self.server, 'mapbox_token':self.mapbox_token})
                if item.get('type') == 'widget' and return_widgets:
                    collection.append({'type': 'Widget', 'id': item.get('id'), 'attributes': item.get('attributes'), 'server': self.server})
        return collection

    def order_results(self, collection_list):
        """Operate on a list of objects given the order key, limit, and rule a user has passed"""
        tmp_sorted = []
        try:
            d = {}
            duplicate = {}
            for n, z in enumerate([c['attributes'].get(self.order.lower()) for c in collection_list]):
                if duplicate.get(z, None):
                    d[f'{z} {duplicate[z]}'] = collection_list[n]
                    duplicate[z] += 1
                else:
                    d[z] = collection_list[n]
                    duplicate[z] = 1
            keys = sorted(d, reverse=self.sort.lower() == 'asc')
            for key in keys:
                tmp_sorted.append(d[key])
        except:
            raise ValueError(f'[Order-error] Param does not exist in collection: {self.order}, rule: {self.sort}')
        if self.limit < len(tmp_sorted):
            tmp_sorted = tmp_sorted[0:self.limit]
        return tmp_sorted

    def save(self, path=None):
        """
        Save all entities in the collection to a local path.
        """
        if not path:
            path = './LMI-BACKUP'
            if not os.path.isdir(path):
                os.mkdir(path)
            today = datetime.datetime.today().strftime('%Y-%m-%d@%Hh-%Mm')
            path += f'/{today}'
            if not os.path.isdir(path):
                os.mkdir(path)
        else:
           if not os.path.isdir(path):
                os.mkdir(path)
        print(f'Saving to path: {path}')
        saved = []
        failed = []
        if server_uses_widgets(self.server):
            url_args = "vocabulary,metadata,layer,widget"
        else:
            url_args = "metadata,layer"
        for item in tqdm(self):
            if item['id'] not in saved:
                entity_type = item.get('type')
                if entity_type in ['Dataset', 'Table']:
                    ds_id = item['id']
                else:
                    ds_id = item['attributes']['dataset']
                try:
                    url = f'{self.server}/v1/dataset/{ds_id}?includes={url_args}'
                    r = requests.get(url)
                    dataset_config = r.json()['data']
                except:
                    failed.append(item)

                save_json = {
                    "id": ds_id,
                    "type": "dataset",
                    "server": self.server,
                    "attributes": dataset_config['attributes']
                }
                with open(f"{path}/{ds_id}.json", 'w') as fp:
                    json.dump(save_json, fp)

        if len(failed) > 0:
            print(f'Some entities failed to save: {failed}')
            return failed
        print('Save complete!')



import requests
import json
import random
import geopandas as gpd
import os
import datetime

#from shapely.geometry import shape
from time import sleep
from pprint import pprint
from .layer import Layer
from .utils import html_box, nested_set, server_uses_widgets
from .lmipy import Vocabulary, Metadata, Widget


class Dataset:
    """
    This is the main Dataset class.

    Parameters
    ----------
    id_hash: int
        An ID hash of the dataset in the API.
    attributes: dic
        A dictionary holding the attributes of a dataset.
    sever: str
        A URL string of the vizzuality server.
    """
    def __init__(self, id_hash=None, attributes=None, server='https://api.resourcewatch.org', token=None):
        self.id = id_hash
        self.layers = []
        self.server = server
        if not attributes:
            self.attributes = self.get_dataset()
        elif attributes and token:
            created_dataset = self.new_dataset(token=token, attributes=attributes, server=server)
            self.attributes = created_dataset.attributes
            self.id = created_dataset.id
        elif attributes:
            self.id = attributes.get('id')
            self.attributes = self.get_dataset()

        if len(self.attributes.get('layer', [])) > 0:
            self.layers = [Layer(id_hash=l.get('id', None), attributes=l, server=self.server) for l in self.attributes.get('layer')]
            _ = self.attributes.pop('layer')
        if len(self.attributes.get('metadata', [])) > 0:
            self.metadata = [Metadata(attributes=m, server=self.server) for m in self.attributes.get('metadata')]
            _ = self.attributes.pop('metadata')
        else:
            self.metadata = []
        if len(self.attributes.get('vocabulary', [])) > 0:
            self.vocabulary =[Vocabulary(attributes=v, server=self.server) for v in self.attributes.get('vocabulary')]
            _ = self.attributes.pop('vocabulary')
        else:
            self.vocabulary = []
        if len(self.attributes.get('widget', [])) > 0:
            self.widget =[Widget(w.get('id'), attributes=1, server=self.server) for w in self.attributes.get('widget')]
            _ = self.attributes.pop('widget')
        else:
            self.widget = []
        self.url = f"{self.server}/v1/dataset/{id_hash}?hash={random.getrandbits(16)}"

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        return f"Dataset {self.id} {self.attributes['name']}"

    def _repr_html_(self):
        return html_box(item=self)

    def get_dataset(self):
        """
        Retrieve a dataset from a server by ID.
        """
        try:
            hash = random.getrandbits(16)
            if server_uses_widgets(self.server):
                url = f'{self.server}/v1/dataset/{self.id}?includes=layer,widget,vocabulary,metadata&hash={hash}'
            else:
                url = f'{self.server}/v1/dataset/{self.id}?includes=layer,metadata&hash={hash}'
            r = requests.get(url)
        except:
            raise ValueError(f'Unable to get Dataset {self.id} from {r.url}')
        if r.status_code == 200:
            return r.json().get('data').get('attributes')
        else:
            raise ValueError(f'Dataset with id={self.id} does not exist.')

    def carto_query(self, sql):
        """
        Returns a GeoPandas GeoDataFrame for CARTO datasets. The sql query should
        always use dataset as the source (i.e. 'from dataset') as this will be
        replaced with the tableName from dataset.attributes.
        """
        sql = sql.lower().replace('from data',f"FROM {self.attributes.get('tableName')}")
        if not self.attributes.get('connectorUrl'):
            raise ValueError("ConnectorUrl attribute missing.")
        account = self.attributes.get('connectorUrl').split('/')[2].split('.')[0]
        urlCarto = f"https://{account}.carto.com/api/v2/sql"
        params = {"q": sql}
        r = requests.get(urlCarto, params=params)
        if r.status_code == 200:
            return gpd.GeoDataFrame(r.json().get('rows'))
        else:
            raise ValueError(f"Bad response from Carto {r.status_code}: {r.json()}")

    def query(self, sql="SELECT * FROM data LIMIT 5"):
        """
        Query a Dataset object

        Returns a table as a from queries against datasets in an API using the query endpoint.
        Table name must be `data` by default (or the CARTO table name, if known)

        Parameters
        ----------
        sql: str
            Valid SQL string.
        """
        provider = self.attributes.get('provider', None)
        if provider == 'cartodb':
            return self.carto_query(sql=sql)
        else:
            raise ValueError(f'Unable to perform query on datasets with provider {provider}. Must be `cartodb`.')

    def head(self, n=5, decode_geom=True, token=None):
        """
        Returns a table as a GeoPandas GeoDataframe from a Vizzuality API using the query endpoint.
        """
        sql = f'SELECT * FROM data LIMIT {n}'
        return self.carto_query(sql=sql)

    def update_keys(self):
        """
        Returns a list of attribute keys which could be updated.
        """
        update_blacklist = ['metadata','layer', 'vocabulary', 'updatedAt', 'userId', 'slug', "clonedHost", "errorMessage", "taskId", "dataLastUpdated"]
        updatable_fields = {f'{k}':v for k,v in self.attributes.items() if k not in update_blacklist}
        uk = list(updatable_fields.keys())
        return uk

    def update(self, update_params=None, token=None, show_difference=False):
        """
        Update a Dataset object

        To view the potential attributes that could be updated use the Dataset.update_keys() method.

        Parameters
        ----------
        update_params: dic
            A dictionary object containing {key: value} pairs of attributes to update.
        token: str
            A valid API key. https://developer.skydipper.com/
        show_difference: bool
            If set to True a verbose description of the updates will be returned to the user.
        """
        if not token:
            raise ValueError(f'[token=None] API TOKEN required for updates.')
        update_blacklist = ['metadata','layer', 'vocabulary', 'updatedAt', 'userId', 'slug', "clonedHost", "errorMessage", "taskId", "dataLastUpdated"]
        attributes = {f'{k}':v for k,v in self.attributes.items() if k not in update_blacklist}
        if not update_params:
            raise ValueError(f'[update_params=None] Must specify update parameters.')
        else:
            payload = {}
            for k, v in update_params.items():
                if '.' in k:
                    nested_keys = k.split('.')
                    if len(nested_keys) > 1 and nested_keys[0] in list(attributes.keys()):
                        payload[nested_keys[0]] = attributes.get(nested_keys[0])
                        nested_set(payload, nested_keys, v)
                elif k in list(attributes.keys()):
                    payload[k] = v
        try:
            url = f"{self.server}/dataset/{self.id}"
            headers = {'Authorization': f'Bearer {token}', 'Content-Type': 'application/json'}
            r = requests.patch(url, data=json.dumps(payload), headers=headers)
        except:
            raise ValueError(f'Dataset update failed.')
        if r.status_code == 200:
            response = r.json()['data']
        else:
            pass
            return None
        self.attributes = self.get_dataset()
        return self

    def confirm_delete(self):
        print(f"Delete Dataset {self.attributes['name']} with id={self.id}?")
        print("Note: Dataset deletion cascades to all associated Layers, Metadata and Vocabularies.\n> y/n")
        conf = input()
        if conf.lower() == 'y':
            return True
        elif conf.lower() == 'n':
            return False
        else:
            print('Requires y/n input!')
            return False

    def delete(self, token=None, force=False):
        """
        Deletes a target Dataset object.
        If force=True, deletes dataset along with all children.
        """
        if not token:
            raise ValueError(f'[token] API token required to delete.')
        layer_count = len(self.layers)
        if layer_count > 0:
            if not force:
                print(f'WARNING - Dataset has {layer_count} associated Layer(s).')
                print('[D]elete ALL associated Layers, or\n[A]bort delete process?')
                conf = input()
            else:
                conf = 'd'
            if conf.lower() == 'd':
                for l in self.layers:
                    l.delete(token, force=True)
            elif conf.lower() == 'a':
                return False
            else:
                print('Requires D/A input!')
                return False
        if not force:
            conf = self.confirm_delete()
        elif force:
            conf = True
        if conf:
            try:
                url = f'{self.server}/dataset/{self.id}'
                headers = {'Authorization': f'Bearer {token}', 'Content-Type': 'application/json', 'Cache-Control': 'no-cache'}
                r = requests.delete(url, headers=headers)
            except:
                raise ValueError(f'Dataset deletion failed.')
            if r.status_code == 200:
                print(r.url)
                print('Deletion successful!')
                self = None
            else:
                raise ValueError(f'Dataset deletion unsuccessful. {r.status_code}')
        else:
            print('Deletion aborted.')
        return self

    def clone(self, token=None, env='staging', clone_server=None, dataset_params=None, clone_children=False):
        """
        Create a clone of a target Dataset as a new staging or prod Dataset.
        A set of attributes can be specified for the clone Dataset.

        The argument `clone_server` specifies the server to clone to. Default server = https://api.resourcewatch.org

        Set clone_children=True to clone all child layers, widgets, vocabs and metadata entities.
        Alternatively you may set clone_children as a list containing one or more or 'layer', 'widget', 'vocab', 'meta'
        in order to selectively clone children by entity type.
        .
        """
        if clone_children == True: clone_children = ['layer', 'widget', 'vocab', 'meta']
        if not clone_server: clone_server = self.server
        if not token:
            raise ValueError(f'[token] API token required to clone.')
        else:
            name = dataset_params.get('name', self.attributes['name'] + 'CLONE')
            clone_dataset_attr = {**self.attributes, 'name': name}
            for k,v in clone_dataset_attr.items():
                if k in dataset_params:
                    clone_dataset_attr[k] = dataset_params.get(k, '')
            payload = {
                'dataset': {
                    'application': clone_dataset_attr['application'],
                    'connectorType': clone_dataset_attr['connectorType'],
                    'connectorUrl': clone_dataset_attr['connectorUrl'],
                    'tableName': clone_dataset_attr['tableName'],
                    'provider': clone_dataset_attr['provider'],
                    'published': clone_dataset_attr['published'],
                    'env': clone_dataset_attr['env'],
                    'name': clone_dataset_attr['name']
                }
            }
            ## wms exception
            if payload['dataset']['connectorType'] == 'wms' and payload['dataset']['tableName'] == None:
                del payload['dataset']['tableName']

            print(f'Creating clone dataset')
            url = f'{clone_server}/dataset'
            headers = {'Authorization': f'Bearer {token}', 'Content-Type': 'application/json', 'Cache-Control': 'no-cache'}
            r = requests.post(url, data=json.dumps(payload), headers=headers)
            if r.status_code == 200:
                clone_dataset_id = r.json()['data']['id']
                clone_dataset = Dataset(id_hash=clone_dataset_id, server=clone_server)
            else:
                print(r.status_code)
                return None
            print(f'{clone_server}/v1/dataset/{clone_dataset_id}')
            if clone_children:
                # Wait for dataset to be added
                sleep(0.5)
                layers =  self.layers
                if len(layers) > 0 and 'layer' in clone_children:
                    for l in layers:
                        try:
                            layer_name = l.attributes['name']

                            l.clone(token=token, env=env, layer_params={'name': layer_name}, clone_server=clone_server, target_dataset_id=clone_dataset_id)
                        except:
                            raise ValueError(f'Layer cloning failed for {l.id}')
                elif len(layers) == 0:
                    print("No child layers to clone!")
                widgets =  self.widget
                if len(widgets) > 0 and 'widget' in clone_children:
                    for w in widgets:
                        widget = w.attributes
                        widget_payload = {
                            "name": widget['name'],
                            "description": widget.get('description', None),
                            "env": payload['dataset']['env'],
                            "widgetConfig": widget['widgetConfig'],
                            "application": payload['dataset']['application']
                        }
                        try:
                            clone_dataset.add_widget(token=token, widget_params=widget_payload)
                        except:
                            raise ValueError(f'Widget cloning failed for {widget.id}')
                elif len(widgets) == 0:
                    print("No child widgets to clone!")
                vocabs = self.vocabulary
                if len(vocabs) > 0 and 'vocab' in clone_children:
                    for v in vocabs:
                        vocab = v.attributes
                        vocab_payload = {
                            'application': vocab['application'],
                            'name': vocab['name'],
                            'tags': vocab['tags']
                        }
                        try:
                            clone_dataset.add_vocabulary(vocab_params=vocab_payload, token=token)
                        except:
                            raise ValueError('Failed to clone Vocabulary.')
                elif len(vocabs) == 0:
                    print("No child vocabs to clone!")
                metas = self.metadata
                if len(metas) > 0 and 'meta' in clone_children:
                    for m in metas:
                        meta = m.attributes
                        meta_payload = {
                            'application': meta['application'],
                            'info': meta['info'],
                            'language': meta['language']
                        }
                        try:
                            clone_dataset.add_metadata(meta_params=meta_payload, token=token)
                        except:
                            raise ValueError('Failed to clone Metadata.')
                elif len(metas) == 0:
                    print("No child metadata to clone!")
            # self.attributes = Dataset(clone_dataset_id, server=clone_server).attributes
            return Dataset(id_hash=clone_dataset_id, server=clone_server)


    def intersect(self, geometry):
        """
        EXPERIMENTAL FEATURE

        Intersect an EE raster with a geometry

        Given a valid LMIPy.Geometry object, return a dictionary based on reduceRegion.

        Parameters
        ---------
        geometry: Geometry
            An LMIPy.Geometry object
        server: str
            A string of a server to call to.
        """
        if self.attributes.get('provider') != 'gee':
            raise ValueError("Intersect currently only supported for EE raster data")
        url = f"{self.server}/query/{self.id}"
        sql = f"SELECT ST_SUMMARYSTATS() from {self.attributes.get('tableName')}"
        params = {"sql": sql,
                  "geostore": geometry.id}
        r = requests.get(url, params=params)
        if r.status_code == 200:
            try:
                return r.json().get('data', [{}])[0].get('st_summarystats', None)
            except:
                raise ValueError(f'Unable to retrieve values from response {r.json()}')
        else:
            print("Hint: sometimes this service fails due to load on EE servers. Try again.")
            raise ValueError(f'Bad response: {r.status_code} from query: {r.url}')

    def save(self, path=None):
        """
        Construct dataset json and save to local path in a date-referenced folder
        """
        if not path:
            path = './LMI-BACKUP'
            if not os.path.isdir(path):
                os.mkdir(path)
            today = datetime.datetime.today().strftime('%Y-%m-%d@%Hh-%Mm')
            path += f'/{today}'
            if not os.path.isdir(path):
                os.mkdir(path)
        else:
           if not os.path.isdir(path):
                os.mkdir(path)
        if server_uses_widgets(self.server):
            url_args = "vocabulary,metadata,layer,widget"
        else:
            url_args = "metadata,layer"
        try:
            url = f"{self.server}/v1/dataset/{self.id}?includes={url_args}"
            r = requests.get(url)
            dataset_config = r.json()['data']
        except:
            raise ValueError(f'Could not retrieve config.')

        save_json = {
            "id": self.id,
            "type": "dataset",
            "server": self.server,
            "attributes": dataset_config['attributes']
        }
        if not os.path.isdir(path):
            os.mkdir(path)
        with open(f"{path}/{self.id}.json", 'w') as fp:
            json.dump(save_json, fp)
        print('Save complete!')


    def load(self, path=None, check=True):
        """
        From a local backup at the specified path, loads and returns a previous version of the current dataset.
        """
        if not path:
            print('Requires a file path to valid backup folder.')
            return None
        try:
            with open(f"{path}/{self.id}.json") as f:
                recovered_dataset = json.load(f)
            server = recovered_dataset.get('server', 'https://api.resourcewatch.org')
            if check:
                blacklist = ['metadata','layer','widget','vocabulary', 'updatedAt']
                attributes = {f'{k}':v for k,v in recovered_dataset['attributes'].items() if k not in blacklist}
                difs = {f'{k}': [v, self.attributes[k]] for k,v in attributes.items() if k not in blacklist and self.attributes[k] != attributes[k]}
                if check and self.attributes == attributes:
                    print('Loaded attributes == existing attributes')
                elif check and self.attributes == attributes:
                    print('Loaded attributes != existing attributes')
                    pprint(difs)
        except:
            raise ValueError(f'Failed to load backup from f{path}/{self.id}.json')
        return Dataset(attributes={**recovered_dataset['attributes'], 'id': recovered_dataset['id']}, server=server)

    def add_vocabulary(self, vocab_params=None, token=None):
        """
        Create a new vocabulary association to the current dataset.

        A single application string, name string and tags list must be specified within the `vocab_params` dictionary.

        A RW-API token is required.
        """
        if not token:
            raise ValueError(f'[token] API token required to create new vocabulary.')
        vocab_type = vocab_params.get('name', None)
        vocab_tags = vocab_params.get('tags', None)
        app = vocab_params.get('application', None)
        ds_id = self.id
        if vocab_tags and len(vocab_tags) > 0 and vocab_type and app:
            payload = {
                "tags": vocab_tags,
                "application": app
            }
            try:
                url = f'{self.server}/v1/dataset/{ds_id}/vocabulary/{vocab_type}'
                headers = {'Authorization': 'Bearer ' + token, 'Content-Type': 'application/json'}
                r = requests.post(url, data=json.dumps(payload), headers=headers)
            except:
                raise ValueError(f'Vocabulary creation failed.')
            if r.status_code == 200:
                print(f'Vocabulary {vocab_type} created.')
                return Dataset(id_hash=ds_id, server=self.server)
            else:
                print(f'Failed with error code {r.status_code}')
                return None
        else:
            raise ValueError(f'Vocabulary creation requires: application string, name string, and a list of tags.')

    def add_metadata(self, meta_params=None, token=None):
        """
        Create a new metadata association to the current dataset.

        A single application string and language string ('en' by default) must be specified within the
        `meta_params` dictionary, as well as an (optional) info dictionary.
        Info has a free schema.

        A RW-API token is required.
        """
        if not token:
            raise ValueError(f'[token] API token required to create new vocabulary.')
        info = meta_params.get('info', None)
        app = meta_params.get('application', None)
        ds_id = self.id
        if info and app:
            payload = {
                "info": info,
                "application": app,
                "language": meta_params.get('language', 'en')
            }
            try:
                url = f'{self.server}/v1/dataset/{ds_id}/metadata'
                headers = {'Authorization': 'Bearer ' + token, 'Content-Type': 'application/json'}
                r = requests.post(url, data=json.dumps(payload), headers=headers)
            except:
                raise ValueError(f'Vocabulary creation failed.')
            if r.status_code == 200:
                print(f'Metadata created.')
                return Dataset(id_hash=ds_id, server=self.server)
            else:
                print(f'Failed with error code {r.status_code}')
                return None
        else:
            raise ValueError(f'Metadata creation requires an info object and application string.')

    def add_widget(self, widget_params=None, token=None):
        """
        Create a new widget association to the current dataset.

        A application list, name and widgetConfig must be specified within the
        `widget_params` dictionary.
        The widgetConfig key has a free schema.

        A RW-API token is required.
        """
        if not token:
            raise ValueError(f'[token] API token required to create new widget.')
        name = widget_params.get('name', None)
        description = widget_params.get('description', None)
        widget_config = widget_params.get('widgetConfig', None)
        app = widget_params.get('application', None)
        ds_id = self.id
        if name and widget_config and app:
            payload = {
                "name": name,
                "description": description,
                "widgetConfig": widget_config,
                "application": app
            }
            try:
                url = f'{self.server}/v1/dataset/{ds_id}/widget'
                print(url)
                headers = {'Authorization': 'Bearer ' + token, 'Content-Type': 'application/json'}
                r = requests.post(url, data=json.dumps(payload), headers=headers)
                print(r.json())
            except:
                raise ValueError(f'Widget creation failed.')
            if r.status_code == 200:
                print(f'Widget created.')
                return Dataset(id_hash=ds_id, server=self.server)
            else:
                print(f'Failed with error code {r.status_code}')
                return None
        else:
            raise ValueError(f'Widget creation requires name string, application list and a widgetConfig object.')

    def new_dataset(self, token=None, attributes=None, server='https://api.resourcewatch.org'):
        """
        Create a new staging or prod Dataset entity from attributes.
        """
        if not token:
            raise ValueError(f'[token] API token required to create a new dataset.')
        elif not attributes:
            raise ValueError(f'Attributes required to create a new dataset.')
        else:
            url = f'{server}/dataset'
            headers = {'Authorization': 'Bearer ' + token, 'Content-Type': 'application/json', 'Cache-Control': 'no-cache'}
            payload = {'dataset': attributes}

            r = requests.post(url, data=json.dumps(payload), headers=headers)
            if r.status_code == 200:
                new_dataset_id = r.json()['data']['id']
            else:
                print(r.status_code)
                return None
            print(f'{self.server}/v1/dataset/{new_dataset_id}')
            return Dataset(id_hash=new_dataset_id, server=server)

    def merge(self, token=None, target_dataset=None, target_dataset_id=None, target_server='https://api.resourcewatch.org', key_whitelist=[], force=False):
        """
        'Merge' one Dataset entity into another target Dataset.
        The argument `key_whitelist` can be used to specify which properties you wish to merge (if not all)
        Note: requires API token.
        """
        if not token:
            raise ValueError(f'[token] API token required to update Dataset.')
        if not target_dataset and target_dataset_id and target_server:
            target_dataset = Dataset(target_dataset_id, server=target_server)
        else:
            raise ValueError(f'Requires either target Dataset or Dataset id plus server.')
        atts = self.attributes
        payload = {
            'connectorType': atts.get('connectorType', None),
            'connectorUrl': atts.get('connectorUrl', None),
            'tableName': atts.get('tableName', None),
            'name': atts.get('name', None),
            'description': atts.get('description', None),
            'application': atts.get('application', None),
            'provider': atts.get('provider', None)
        }
        if not key_whitelist: key_whitelist = [k for k in payload.keys()]
        filtered_payload = {k:v for k,v in payload.items() if v and k in key_whitelist}
        print(f'Merging {self.id} from {self.server} into {target_dataset_id} on {target_server}.\nAre you sure you sure you want to continue?')
        if not force:
            conf = input()
        else:
            conf = 'y'
        if conf.lower() == 'y':
            try:
                merged_dataset = target_dataset.update(update_params=filtered_payload, token=token)
            except:
                print('Aborting...')
            print('Completed!')
            return merged_dataset

        elif conf.lower() == 'n':
            print('Aborting...')
            return False
        else:
            print('Requires y/n input!')
            return False



import requests
import geopandas as gpd
import folium
import urllib
import json
import random
import re
from pprint import pprint
from .utils import html_box, get_geojson_string, nested_set, server_uses_widgets


class Layer:
    """
    This is the main Layer class.

    Parameters
    ----------
    id_hash: int
        An ID hash.
    attributes: dic
        A dictionary holding the attributes of a dataset.
    server: str
        A string of the server URL.
    """
    def __init__(self, id_hash=None, attributes=None,
                    server='https://api.resourcewatch.org', mapbox_token=None, token=None):
        self.server = server
        self.mapbox_token = mapbox_token
        if not attributes and id_hash:
            self.id = id_hash
            self.attributes = self.get_layer()
        elif attributes and token:
            created_layer = self.new_layer(token=token, attributes=attributes, server=self.server)
            self.attributes = created_layer.attributes
            self.id = created_layer.id
        elif attributes:
            self.id = attributes.get('id')
            self.attributes = self.get_layer()

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        return f"Layer {self.id} {self.attributes['name']}"

    def _repr_html_(self):
        return html_box(item=self)

    def get_layer(self):
        """
        Returns a layer from a Vizzuality API.
        """
        try:
            hash = random.getrandbits(16)
            if server_uses_widgets(self.server):
                url = f'{self.server}/v1/layer/{self.id}?hash={hash}'
            else:
                url = f'{self.server}/v1/layer/{self.id}?hash={hash}'
            r = requests.get(url)
        except:
            raise ValueError(f'Unable to get Layer {self.id} from {r.url}')
        if r.status_code == 200:
            return r.json().get('data').get('attributes')
        else:
            raise ValueError(f'Layer with id={self.id} does not exist for server={self.server}.')

    def parse_map_url(self):
        """
        Parses map urls
        """
        if self.attributes.get('layerConfig') == None:
            raise ValueError("No layerConfig present in layer from which to create a map.")
        # if tileLayer
        if self.attributes.get('provider') == 'leaflet' and self.attributes.get('layerConfig').get('type') == 'tileLayer':
            return self.get_leaflet_tiles()
        # if GEE
        if self.attributes.get('provider') == 'gee':
            return self.get_ee_tiles()
        # If CARTO
        if self.attributes.get('provider') == 'cartodb':
            return self.get_carto_tiles()
        if self.attributes.get('provider') == 'mapbox':
            if not self.mapbox_token:
                raise ValueError("Requires a Mapbox Access Token in param: 'mapbox_token'.")
            return self.get_mapbox_tiles()

    def get_leaflet_tiles(self):
        """
        Returns leaflet urls.
        """
        url = self.attributes.get('layerConfig').get('url', None)
        if not url:
            url = self.attributes.get('layerConfig').get('body').get('url')
        # This below code is an issue. Not working and probably wont fix the problem
        # as far as I can see. E.g. check Forma case. tmp hack for now is to catch directly.
        # params_config = self.attributes.get('layerConfig').get('params_config', None)
        # if params_config:
        #     for config in params_config:
        #         key = config['key']
        #         default = config['default']
        #         required = config['required']
        #         if required:
        #             url = url.replace(f'{{{key}}}', f'{default}')
        if '{thresh}' in url:
            # try to replace thresh with a best-guess valid threshold (say 30%)
            url = url.replace('{thresh}','30')
        if '{{date}}' in url:
            # try to replace date with a best-guess valid date
            url = url.replace('{{date}}','20190331')
        return url

    def get_ee_tiles(self):
        """Returns tiles from EE assets"""
        url = f'{self.server}/v1/layer/{self.id}/tile/gee/{{z}}/{{x}}/{{y}}'
        return url

    def get_carto_tiles(self):
        """Get carto tiles"""
        sql_config = self.attributes.get('layerConfig').get('sql_config', None)
        layerConfig = self.attributes.get('layerConfig')
        if sql_config:
            for config in sql_config:
                key = config['key']
                key_params = config['key_params']
                if key_params[0].get('required', False):
                    for l in layerConfig["body"]["layers"]:
                        l['options']['sql'] = l['options']['sql'].replace(f'{{{key}}}', '0').format(key_params['key'])
                else:
                    for l in layerConfig["body"]["layers"]:
                        l['options']['sql'] = l['options']['sql'].replace(f'{{{key}}}', '0').format('')
        _layerTpl = urllib.parse.quote_plus(json.dumps({
            "version": "1.3.0",
            "stat_tag": "API",
            "layers": [{ **l, "options": { **l["options"]}} for l in layerConfig.get("body").get("layers")]
        }))
        apiParams = f"?stat_tag=API&config={_layerTpl}"
        url = f"https://{layerConfig.get('account')}.carto.com/api/v1/map{apiParams}"
        r = requests.get(url, headers={'Content-Type': 'application/json'})
        if r.status_code == 200:
            response = r.json()
        else:
            raise ValueError(f'Unable to get retrieve map url for {self.id} from {self.attributes.get("provider")}')

        tile_url = f'{response["cdn_url"]["templates"]["https"]["url"]}/{layerConfig["account"]}/api/v1/map/{response["layergroupid"]}/{{z}}/{{x}}/{{y}}.png'
        return tile_url

    def get_mapbox_tiles(self):
        """"Retrieve mapbox tiles... as raster :("""
        layerConfig = self.attributes['layerConfig']
        vector_target = layerConfig['body'].get('format', None)
        if vector_target and vector_target.lower() == 'mapbox':
            vector_source = layerConfig['body'].get('url', '').split('mapbox://')[1]
            url = f"https://api.mapbox.com/v4/{vector_source}.json?secure&access_token={self.mapbox_token}"
            r = requests.get(url, headers={'Content-Type': 'application/json'})
            if r.status_code == 200:
                return r.json().get('tiles', [None])[0].replace('vector.pbf', 'png')
            else:
                raise ValueError(f'Unable to get retrieve map url for {self.id} from {vector_target.lower()}')
        else:
            raise ValueError('Mapbox target not found')

    def map(self, lat=0, lon=0, zoom=3, geometry=None, color='#64D1B8', weight=4):
        """
        Returns a folim map with styles applied.

        Parameters
        ----------
        lat: float
            A latitude to focus the map on.
        lon: float
            A longitude to focus the map on.
        zoom: int
            A z-level for the map.
        geometry: LMIPy.Geometry()
            A geometry object.
        weight: int
            Weight of geom outline. Default = 4.
        color: str
            Hex code for geom outline. Default = #64D1B8.
        """
        url = self.parse_map_url()
        map = folium.Map(
                location=[lon, lat],
                zoom_start=zoom,
                tiles='OpenStreetMap',
                detect_retina=True,
                prefer_canvas=True
        )
        map.add_tile_layer(tiles=url, attr=self.attributes.get('name'))
        if geometry:
            geojson = geometry.attributes['geojson']
            geom = geojson['features'][0]['geometry']
            centroid = list(geometry.shape()[0].centroid.coords)
            bounds = [geometry.attributes['bbox'][2:][::-1],
                      geometry.attributes['bbox'][:2][::-1]]
            if geom['type'] == 'Point':
                folium.Marker(
                    centroid
                ).add_to(map)
            else:
                folium.GeoJson(
                    data=get_geojson_string(geom),
                    style_function=lambda x: {
                    'fillOpacity': 0,
                    'weight': weight,
                    'color': color
                    }
                ).add_to(map)
            map.fit_bounds(bounds)
        return map

    def update_keys(self):
        """
        Returns a list of theattribute values which could be updated
        """
        update_blacklist = ['updatedAt', 'userId', 'dataset', 'slug']
        updatable_fields = {f'{k}':v for k,v in self.attributes.items() if k not in update_blacklist}
        uk = list(updatable_fields.keys())
        return uk

    def update(self, update_params=None, token=None):
        """
        Update layer specific attribute values

        Pass a dictionary of update_params and update a layer target.

        Parameters
        ----------
        update_params: dic
            A dictionary of update paramters. You can identify the possible keys by calling
            self.update_keys(silent=False)
        token: str
            A valid API Token.
        """
        if not token:
            raise ValueError(f'[token=None] API TOKEN required for updates.')
        update_blacklist = ['updatedAt', 'userId', 'dataset', 'slug']
        attributes = {f'{k}':v for k,v in self.attributes.items() if k not in update_blacklist}
        if not update_params:
            raise ValueError(f'[update_params=None] Must specify update parameters.')
        else:
            payload = {}
            for k, v in update_params.items():
                if '.' in k:
                    nested_keys = k.split('.')
                    if len(nested_keys) > 1 and nested_keys[0] in list(attributes.keys()):
                        payload[nested_keys[0]] = attributes.get(nested_keys[0])
                        nested_set(payload, nested_keys, v)
                elif k in list(attributes.keys()):
                    payload[k] = v
        try:
            url = f"{self.server}/dataset/{self.attributes['dataset']}/layer/{self.id}"
            headers = {'Authorization': f'Bearer {token}', 'Content-Type': 'application/json'}
            r = requests.patch(url, data=json.dumps(payload), headers=headers)
        except:
            raise ValueError(f'Layer update failed.')
        if r.status_code == 200:
            response = r.json()['data']
        else:
            print(f"PATCH attempt threw a {r.status_code}!")
            return None
        self.attributes = self.get_layer()
        return self

    def confirm_delete(self):
        print(f"Delete Layer {self.attributes['name']} with id={self.id}?\n> y/n")
        conf = input()
        if conf.lower() == 'y':
            return True
        elif conf.lower() == 'n':
            return False
        else:
            print('Requires y/n input!')
            return False

    def delete(self, token=None, force=False):
        """
        Deletes a target layer
        """
        if not token:
            raise ValueError(f'[token=None] API token required to delete.')
        if not force:
            conf = self.confirm_delete()
        elif force:
            conf = True
        if conf:
            try:
                url = f'{self.server}/dataset/{self.attributes["dataset"]}/layer/{self.id}'
                headers = {'Authorization': f'Bearer {token}', 'Content-Type': 'application/json', 'Cache-Control': 'no-cache'}
                r = requests.delete(url, headers=headers)
            except:
                raise ValueError(f'Layer deletion failed.')
            if r.status_code == 200:
                print(r.url)
                print('Deletion successful!')
                self = None
        else:
            print('Deletion aborted')
        return None

    def clone(self, token=None, env='staging', clone_server=None, layer_params={}, target_dataset_id=None):
        """
        Create a clone of current Layer (and its parent Dataset) as a new staging or prod Layer.
        A set of attributes can be specified for the clone Layer using layer_params.
        Optionally, you can also select a target Dataset to attach your Layer clone to.

        The argument `clone_server` specifies the server to clone to. Default server is the layers own server.
        """
        from .dataset import Dataset
        if not clone_server: clone_server = self.server

        if not token:
            raise ValueError(f'[token] API token required to clone.')
        target_layer_name  = self.attributes['name']
        name = layer_params.get('name', f'{target_layer_name} CLONE')
        clone_layer_attr = {**self.attributes, 'name': name}
        for k in clone_layer_attr.keys():
            if k in layer_params:
                clone_layer_attr[k] = layer_params[k]
        if target_dataset_id:
            target_dataset = Dataset(id_hash=target_dataset_id, server=clone_server)
        else:
            target_dataset = self.dataset()
            clone_dataset_attr = {**target_dataset.attributes, 'name': name, }
            payload = {"dataset":{
                'application': clone_dataset_attr['application'],
                'connectorType': clone_dataset_attr['connectorType'],
                'connectorUrl': clone_dataset_attr['connectorUrl'],
                'published': clone_dataset_attr['published'],
                'tableName': clone_dataset_attr['tableName'],
                'provider': clone_dataset_attr['provider'],
                'env': env,
                'name': clone_dataset_attr['name']
                }
            }
            print(f'Creating clone dataset')
            url = f'{clone_server}/dataset'
            headers = {'Authorization': f'Bearer {token}', 'Content-Type': 'application/json', 'Cache-Control': 'no-cache'}
            r = requests.post(url, data=json.dumps(payload), headers=headers)
            print(r.url)
            pprint(payload)
            if r.status_code == 200:
                target_dataset_id = r.json()['data']['id']
            else:
                print(r.status_code)
                return None

        payload = {
            'application': clone_layer_attr['application'],
            'applicationConfig': clone_layer_attr['applicationConfig'],
            'description': clone_layer_attr.get('description', ''),
            'env': env,
            'interactionConfig': clone_layer_attr['interactionConfig'],
            'iso': clone_layer_attr['iso'],
            'layerConfig': clone_layer_attr['layerConfig'],
            'legendConfig':clone_layer_attr['legendConfig'] ,
            'name': name,
            'provider': clone_layer_attr['provider'],
            'published': clone_layer_attr['published']
        }
        print(f'Creating clone layer on target dataset')
        url = f'{clone_server}/dataset/{target_dataset_id}/layer'
        headers = {'Authorization': f'Bearer {token}', 'Content-Type': 'application/json', 'Cache-Control': 'no-cache'}
        r = requests.post(url, data=json.dumps(payload), headers=headers)
        if r.status_code == 200:
                clone_layer_id = r.json()['data']['id']
        else:
            print(r.status_code)
            return None
        print(f'{clone_server}/v1/dataset/{target_dataset_id}/layer/{clone_layer_id}')
        return Layer(id_hash=clone_layer_id, server=clone_server)

    def parse_query(self, sql):
        """
        Distributer to decide interect method
        """
        if self.attributes.get('layerConfig') == None:
            raise ValueError("No layerConfig present in layer from which to create a query.")
        # if tileLayer
        if self.attributes.get('provider') == 'leaflet' and self.attributes.get('layerConfig').get('type') == 'tileLayer':
            print(f"Queries on provider type {self.attributes.get('provider')} currently unavailable.")
            return None
        # if GEE
        if self.attributes.get('provider') == 'gee':
            print(f"Queries on provider type {self.attributes.get('provider')} currently unavailable.")
            return None
        # If CARTO
        if self.attributes.get('provider') == 'cartodb':
            return self.get_carto_query(sql)
        return None

    def get_carto_query(self, sql):
        """
        Intersect layer against some geometry class object, geosjon object, shapely shape, or by id.
        """
        attributes = self.attributes
        sql_config = attributes.get('layerConfig').get('sql_config', None)
        layerConfig = attributes.get('layerConfig')
        if sql_config:
            for config in sql_config:
                key = config['key']
                key_params = config['key_params']
                if key_params[0].get('required', False):
                    for l in layerConfig["body"]["layers"]:
                        l['options']['sql'] = l['options']['sql'].replace(f'{{{key}}}', '0').format(key_params['key'])
                else:
                    for l in layerConfig["body"]["layers"]:
                        l['options']['sql'] = l['options']['sql'].replace(f'{{{key}}}', '0').format('')
        base_query = f'with t as ({layerConfig["body"]["layers"][0]["options"]["sql"]}) '
        sql = re.sub("from data", "from t", sql, flags=re.I)
        sql = base_query + sql.replace('"', "'")
        account = layerConfig.get('account')
        urlCarto = f"https://{account}.carto.com/api/v2/sql"
        params = {"q": sql}
        r = requests.get(urlCarto, params=params)
        if r.status_code == 200:
            return gpd.GeoDataFrame(r.json().get('rows'))
        else:
            print(f'{r.url}')
            raise ValueError(f"Bad response from Carto {r.status_code}: {r.json()}")

    def query(self, sql='SELECT * FROM data LIMIT 5'):
        """
        Intersect layer against some geometry class object, geosjon object, shapely shape, or by id.
        """
        return self.parse_query(sql=sql)

    def dataset(self):
        """
        Returns parent datset
        """
        from .dataset import Dataset
        return Dataset(self.attributes['dataset'], server=self.server)

    def intersect(self, geometry):
        """
        Intersect an EE raster with a geometry

        Given a valid LMIPy.Geometry object, return a dictionary based on reduceRegion
        Parameters
        ---------
        geometry: Geometry
            An LMIPy.Geometry object
        server: str
            A string of a server to call to.
        """
        if self.attributes.get('provider') != 'gee':
            raise ValueError("Intersect currently only supported for EE raster data")
        url = f"{self.server}/query/{self.attributes.get('dataset')}"
        sql = f"SELECT ST_SUMMARYSTATS() from {self.attributes.get('layerConfig').get('assetId')}"
        params = {"sql": sql,
                  "geostore": geometry.id}
        r = requests.get(url, params=params)
        if r.status_code == 200:
            try:
                return r.json().get('data', None)[0].get('st_summarystats')
            except:
                raise ValueError(f'Unable to retrieve values from response {r.json()}')
        else:
            print("Hint: sometimes this service fails due to load on EE servers. Try again.")
            raise ValueError(f'Bad response: {r.status_code} from query: {r.url}')

    def save(self, path=None):
        """
        Construct dataset json and save to local path in a date-referenced folder
        """
        from .dataset import Dataset
        self.dataset().save(path=path)

    def load(self, path=None, check=True):
        """
        From a local backup at the specified path, loads and returns a previous version of the current dataset.
        """
        from .dataset import Dataset
        if not path:
            print('Requires a file path to valid backup .json file.')
            return None
        try:
            ds = self.dataset()
            with open(f"{path}/{ds.id}.json") as f:
                recovered_dataset = json.load(f)
            layers = recovered_dataset['attributes']['layer']
            server = recovered_dataset.get('server', 'https://api.resourcewatch.org')
            if len(layers) > 0: recovered_layer = [l for l in layers if l['id'] == self.id][0]
            else: raise ValueError(f'No save layers found!')
            if check:
                blacklist = ['updatedAt']
                attributes = {f'{k}':v for k,v in recovered_layer['attributes'].items() if k not in blacklist}
                if self.attributes == attributes:
                    print('Loaded == existing')
                elif check:
                    print('Loaded != existing')
        except:
            raise ValueError(f'Failed to load backup from f{path}')

        return Layer(attributes={**recovered_layer['attributes'], 'id': recovered_layer['id']}, server=server)


    def new_layer(self, token=None, attributes=None, server='https://api.resourcewatch.org'):
        """
        Create a new staging or prod Layer entity from attributes.
        """
        if not token:
            raise ValueError(f'[token] API token required to create a new dataset.')
        elif not attributes:
            raise ValueError(f'Attributes required to create a new dataset.')
        elif not attributes.get('dataset', None):
            raise ValueError(f'Attributes must include dataset key.')
        else:
            dataset_id = attributes['dataset']
            url = f'{server}/v1/dataset/{dataset_id}/layer'
            headers = {'Authorization': 'Bearer ' + token, 'Content-Type': 'application/json', 'Cache-Control': 'no-cache'}
            payload = {**attributes}

            r = requests.post(url, data=json.dumps(payload), headers=headers)
            if r.status_code == 200:
                new_layer_id = r.json()['data']['id']
            else:
                print(r.status_code)
                return None
            print(f'{server}/v1/layer/{new_layer_id}')
            return Layer(id_hash=new_layer_id, server=server)

    def merge(self, token=None, target_layer=None, target_layer_id=None, target_server='https://api.resourcewatch.org', key_whitelist=[], force=False):
        """
        'Merge' one Layer entity into another target Layer.
        The argument `key_whitelist` can be used to specify which properties you wish to merge (if not all)
        Note: requires API token.
        """
        if not token:
            raise ValueError(f'[token] API token required update Layer.')
        if not target_layer and target_layer_id and target_server:
            target_layer = Layer(target_layer_id, server=target_server)
        elif not target_layer:
            raise ValueError(f'Requires either target Layer or Layer id plus server.')
        atts = self.attributes
        payload = {
            'layerConfig': atts.get('layerConfig', None),
            'legendConfig': atts.get('legendConfig', None),
            'applicationConfig': atts.get('applicationConfig', None),
            'interactionConfig': atts.get('interactionConfig', None),
            'name': atts.get('name', None),
            'description': atts.get('description', None),
            'iso': atts.get('iso', None),
            'application': atts.get('application', None),
            'provider': atts.get('provider', None)
        }
        if not key_whitelist: key_whitelist = [k for k in payload.keys()]
        filtered_payload = {k:v for k,v in payload.items() if v and k in key_whitelist}
        print(f'Merging {self.id} from {self.server} into {target_layer.id} on {target_layer.server}.\nAre you sure you sure you want to continue? Y/N')
        if not force:
            conf = input()
        else:
            conf = 'y'
        if conf.lower() == 'y':
            try:
                merged_layer = target_layer.update(update_params=filtered_payload, token=token)
            except:
                print('Aborting...')
            print('Completed!')
            return merged_layer

        elif conf.lower() == 'n':
            print('Aborting...')
            return False
        else:
            print('Requires y/n input!')
            return False



from .utils import html_box, get_geojson_string
import requests
import json
import folium
import numpy as np
import random
import png
import geopandas as gpd
from shapely.geometry.polygon import LinearRing
from shapely.geometry import shape

class Image:
    """
    Main Image Class

    Parameters
    ----------
    bbox: dict
        A dictionary describing the bounding box of the image.

    in: float
        A decimal longitude.

    bands: list
        A list of bands to visulise (e.g. ['b4','b3','b2']).

    instrument: str
        A string indicating the satellite platform ('sentinel', 'landsat', 'all').

    start: str
        Start date ('YYYY-MM-DD') to bound the search for the satellite images.

    end: str
        End date ('YYYY-MM-DD') to bound the search for the satellite images.

    """

    def __init__(self, source=None, instrument=None, date_time=None, cloud_score=None,
                 thumb_url = None, bbox=None, tile_url=None, ring=None,
                 server='http://api.resourcewatch.org', type=None, np_array=None,
                 np_array_bounds=None, band_viz={'bands': ['B4', 'B3', 'B2'], 'min': 0, 'max': 0.4}):
        self.source = source
        if type == None:
            self.type = 'Image'
        else:
            self.type = type
        self.instrument = instrument
        self.cloud_score = cloud_score
        self.date_time = date_time
        self.server = server
        self.band_viz = band_viz
        self.bbox = bbox
        if ring == None:
            self.ring = self.get_ring()
        else:
            self.ring = ring
        self.tile_url = tile_url
        if thumb_url:
            self.thumb_url = thumb_url
        else:
            self.thumb_url = self.get_thumbs()
        self.attributes = self.get_attributes()
        self.np_array = np_array
        self.np_array_bounds = np_array_bounds

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        return f"Image {self.source}"

    def _repr_html_(self):
        return html_box(item=self)

    def get_thumbs(self):
        payload = {'source_data': [{'source': self.source}], 'bands': self.band_viz.get('bands')}
        url = self.server + '/recent-tiles/thumbs'
        r = requests.post(url, data=json.dumps(payload), headers={'Content-Type': 'application/json'})
        if  r.status_code == 200:
            return r.json().get('data').get('attributes')[0].get('thumbnail_url')
        else:
            print(f'Failed to get tile {r.status_code}, {r.json()}')
            return None

    def get_image_url(self):
        payload = {'source_data': [{'source': self.source}], 'bands': self.band_viz.get('bands')}
        url = self.server + '/recent-tiles/tiles'
        r = requests.post(url, data=json.dumps(payload), headers={'Content-Type': 'application/json'})
        if  r.status_code == 200:
            return r.json().get('data').get('attributes')[0].get('tile_url')
        else:
            print(f'Failed to get tile {r.status_code}, {r.json()}')
            return None

    def get_ring(self):
            coords = self.bbox.get('geometry').get('coordinates', None)
            if coords and any(isinstance(i, list) for i in coords[0]):
                coords = coords[0]
            ring = LinearRing(coords)
            return ring

    def get_attributes(self):
        return {'provider': self.source}

    def map(self, color='#64D1B8', weight=6):
        """
        Returns a folim map of selected image with styles applied.

        Parameters
        ----------
        weight: int
            Weight of geom outline. Default = 6.
        color: str
            Hex code for geom outline. Default = #64D1B8.
        """
        centroid = [self.ring.centroid.xy[1][0], self.ring.centroid.xy[0][0]]
        result_map = folium.Map(location=centroid, tiles='OpenStreetMap')
        geojson_str = get_geojson_string(self.bbox['geometry'])
        folium.GeoJson(
                    data=geojson_str,
                    style_function=lambda x: {
                    'fillOpacity': 0,
                    'weight': weight,
                    'color': color
                    }
                ).add_to(result_map)
        w,s,e,n = list(self.ring.bounds)
        result_map.fit_bounds([[s, w], [n, e]])
        if self.np_array is not None:
            blist = [[self.np_array_bounds[1], self.np_array_bounds[0]],[self.np_array_bounds[3], self.np_array_bounds[2]]]
            image_overlay = folium.raster_layers.ImageOverlay(self.np_array, blist)
            result_map.add_child(image_overlay)
            return result_map
        elif not self.tile_url:
            self.tile_url = self.get_image_url()
        if self.tile_url:
            result_map.add_tile_layer(tiles=self.tile_url, attr=f"{self.instrument} image")
            return result_map

    def classify(self, model_type='random_forest', version=None):
        """
        Classify an Image Object with a selection of pre-trained models. Returns an Image Object.
        Models avaiable for use are Random Forest, Segnet, and Deepvel.

        Parameters
        ----------
        model_type: string
            A string ('random_forest', 'segnet', or 'deepvel') determining which type of classification will be done.

        version: string
            A string specifying the version of the model to use (e.g. 'v1', 'v2', 'v3'). If not provided, the latest
            version of the model will be used.
        """
        if model_type == 'random_forest':
            if self.type in ['Composite Image', 'Classified Image']:
                raise ValueError(f'Unable to perform {model_type} classification on a {self.type}.')
            url = self.server + '/recent-tiles-classifier'
            params = {'img_id': self.attributes.get('provider')}
            r = requests.get(url, params=params)
            if r.status_code == 200:
                classified_tiles = r.json().get('data').get('attributes').get('url')
                tmp = {'instrument': self.instrument,
                        'date_time': self.date_time,
                        'cloud_score': self.cloud_score,
                        'source': self.source,
                        'band_viz': None,
                        'ring': self.ring,
                        'server': self.server,
                        'thumb_url': self.thumb_url,
                        'tile_url': classified_tiles,
                        'type': 'Classified Image',
                        'bbox': self.bbox,
                        'np_array_bounds': self.np_array_bounds
                        }
                return Image(**tmp)
            else:
                raise ValueError(f'Classification failed ({r.status_code} response): {r.json()}')
            return None
        if model_type in ['segnet', 'deepvel']: #and self.type == 'Composite Image':
            if self.type in ['Classified Image']:
                raise ValueError(f"Unable to perform {model_type} classification on a {self.type}.")
            payload = {'thumb_url': self.thumb_url,
                        'model_name': 'deepvel',
                        'model_version': None}
            url = f'https://us-central1-skydipper-196010.cloudfunctions.net/classify'
            headers = {'Content-Type': 'application/json'}
            r = requests.post(url, data=json.dumps(payload), headers=headers)
            if r.status_code == 200:
                image = np.array(r.json().get('output'), dtype=np.uint8)
                hash_code = random.getrandbits(128)
                thumb_path = f"./{str(hash_code)[0:5]}.png"
                p = png.from_array(image, mode='RGB').save(thumb_path)
                tmp = {'instrument': self.instrument,
                    'date_time': self.date_time,
                    'cloud_score': self.cloud_score,
                    'source': self.source,
                    'band_viz': None,
                    'ring': self.ring,
                    'server': self.server,
                    'thumb_url': thumb_path,
                    'tile_url': None,
                    'type': 'Classified Image',
                    'bbox': self.bbox,
                    'np_array': image,
                    'np_array_bounds': self.np_array_bounds
                    }
                return Image(**tmp)
            else:
                raise ValueError(f"Classification service responded with {r.status_code}: {r.url}")
        else:
            raise ValueError(f"Model type {model_type} not reccognised. type property should be one of 'random_forest', 'segnet', or 'deepvel'")